@@ -1,19 +1,19 @@
 /* Test for Pthreads/mutexes.
-   Copyright (C) 2000 Free Software Foundation, Inc.
+   Copyright (C) 2000, 2001, 2002 Free Software Foundation, Inc.
    This file is part of the GNU C Library.
    Contributed by Kurt Garloff <garloff@suse.de>, 2000.
 
    The GNU C Library is free software; you can redistribute it and/or
-   modify it under the terms of the GNU Library General Public License as
-   published by the Free Software Foundation; either version 2 of the
+   modify it under the terms of the GNU Lesser General Public License as
+   published by the Free Software Foundation; either version 2.1 of the
    License, or (at your option) any later version.
 
    The GNU C Library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   Library General Public License for more details.
+   Lesser General Public License for more details.
 
-   You should have received a copy of the GNU Library General Public
+   You should have received a copy of the GNU Lesser General Public
    License along with the GNU C Library; see the file COPYING.LIB.  If not,
    write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.  */
@@ -25,6 +25,9 @@
 #include <string.h>
 #include <unistd.h>
 
+static void *thread_start (void *ptr) __attribute__ ((__noreturn__));
+
+
 struct thr_ctrl
 {
   pthread_mutex_t mutex;
@@ -35,8 +38,7 @@ struct thr_ctrl
 static void
 dump_mut (pthread_mutex_t * mut)
 {
-  int i;
-
+  size_t i;
   for (i = 0; i < sizeof (*mut); i++)
     printf (" %02x", *((unsigned char *) mut + i));
   printf ("\n");
@@ -58,7 +60,6 @@ pthr_cond_signal_mutex (pthread_cond_t * cond, pthread_mutex_t * mut)
     printf ("mutex_unlock: %s\n", strerror (err));
 }
 
-
 static void *
 thread_start (void *ptr)
 {
